﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Threading;

namespace AControler {
	/// <summary>
	/// Adventurer3との通信用制御クラス
	/// </summary>
	public class Adventurer3
    {
		#region ARPテーブル取り出し用DLL
		[DllImport("iphlpapi.dll")]
		extern static int GetIpNetTable(IntPtr pTcpTable, ref int pdwSize, bool bOrder);

		[StructLayout(LayoutKind.Sequential)]
		public struct MIB_IPNETROW {
			public int Index;
			public int PhysAddrLen;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
			public byte[] PhysAddr;
			public int Addr;
			public int Type;
		}
		#endregion

		/// <summary>
		/// Adventurer3のIPをサーチするクラス
		/// ARPテーブルから登録済みIPを検索し、そのIPアドレスのMACアドレス先頭02で始まるものを返す
		/// ARPテーブル検索は、GetIpNetTableを使用する
		/// </summary>
		public static List<IPAddress> searchIP() {
			var ips = new List<IPAddress>();

			int size = 0;
			GetIpNetTable(IntPtr.Zero, ref size, true);//必要サイズの取得
			var p = Marshal.AllocHGlobal(size);//メモリ割当て
			if (GetIpNetTable(p, ref size, true) == 0) {//データの取得
				var num = Marshal.ReadInt32(p);//MIB_IPNETTABLE.dwNumEntries(データ数)
				var ptr = IntPtr.Add(p, 4);
				for (int i = 0; i < num; i++) {
					MIB_IPNETROW work = (MIB_IPNETROW)Marshal.PtrToStructure(ptr, typeof(MIB_IPNETROW));
					if (work.PhysAddrLen > 0 && work.PhysAddr[0] == 0x02 && work.Type == 3) {
						ips.Add(new IPAddress(BitConverter.GetBytes(work.Addr)));
					}
					ptr = IntPtr.Add(ptr, Marshal.SizeOf(typeof(MIB_IPNETROW)));//次のデータ
				}
				Marshal.FreeHGlobal(p);  //メモリ開放
			}

			return ips;
		}

		/// <summary>
		/// Adventurer3への接続ポート番号(一応固定)
		/// </summary>
		const int Adventurer3Port = 8899;
		string ip_;
		System.Net.Sockets.TcpClient tcp_;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="ip">接続先機器のIPアドレス</param>
		public Adventurer3(IPAddress ip) {
			ip_ = ip.ToString();
		}

		/// <summary>
		/// 接続先IP情報の取り出し
		/// </summary>
		public string ip {
			get { return ip_; }
		}

		/// <summary>
		/// データ受信処理
		/// </summary>
		/// <param name="ns">TCPストリーム</param>
		/// <returns>受信文字列</returns>
		private string recv(NetworkStream ns) {
			System.Text.Encoding enc = System.Text.Encoding.UTF8;
			// サーバーから送られたデータを受信する
			string ret = "";
			var resBytes = new byte[256];
			int resSize = 0;
			do {
				resSize = ns.Read(resBytes, 0, resBytes.Length);
				if (resSize == 0) {
					break;
				}
				string work = new string(enc.GetChars(resBytes), 0, resSize);
				ret += work;
			} while (ns.DataAvailable);
			return ret;
		}

		/// <summary>
		/// 機器と接続状態かどうかの判断
		/// </summary>
		public bool isConnected {
			get {
				if (tcp_ == null) {
					return false;
				}
				else if (tcp_.Connected) {
					return true;
				}
				else {
					tcp_.Close();
					tcp_ = null;
					return false;
				}
			}
		}

		/// <summary>
		/// 接続開始処理
		/// </summary>
		/// <returns>開始できた</returns>
		public bool start() {
			if (tcp_ == null) {
				try {
					tcp_ = new TcpClient(ip_, Adventurer3Port);
				}
				catch(SocketException) {
					// ソケット接続エラーの場合
					return false;
				}
				tcp_.ReceiveTimeout = 5000;
				var ns = tcp_.GetStream();
				System.Text.Encoding enc = System.Text.Encoding.UTF8;
				byte[] sendBytes = enc.GetBytes("~M601 S1\r\n");
				//データを送信する
				ns.Write(sendBytes, 0, sendBytes.Length);
				recv(ns);
			}
			return true;
		}

		/// <summary>
		/// 接続終了処理
		/// </summary>
		public void end() {
			if (tcp_ != null) {
				var ns = tcp_.GetStream();
				System.Text.Encoding enc = System.Text.Encoding.UTF8;
				byte[] sendBytes = enc.GetBytes("~M602\r\n");
				//データを送信する
				ns.Write(sendBytes, 0, sendBytes.Length);
				recv(ns);
				tcp_.Close();
				tcp_ = null;
			}
		}

		/// <summary>
		/// コマンド送信処理
		/// </summary>
		/// <param name="cmd">コマンド文字列</param>
		/// <param name="preOut">コマンドの前に~をつけるかつけないか</param>
		/// <returns>コマンド送信後の受信文字列</returns>
		public string send(string cmd, bool preOut = true) {
			System.Text.Encoding enc = System.Text.Encoding.UTF8;
			byte[] sendBytes = enc.GetBytes(string.Format("{1}{0}\r\n", cmd, preOut ? "~" : ""));
			return send(sendBytes);
		}

		/// <summary>
		/// データ送信処理
		/// </summary>
		/// <param name="data">送信データ</param>
		/// <returns>受信文字列</returns>
		public string send(byte[] data) {
			if (tcp_ != null) {
				var ns = tcp_.GetStream();
				//データを送信する
				try {
					ns.Write(data, 0, data.Length);
					return recv(ns);
				}
				catch (IOException) {
					// IO例外があった場合、とりあえずtcpを閉じてしまう。
					tcp_.Close();
					tcp_ = null;
					return null;
				}
			}
			else {
				return null;
			}
		}


		/// <summary>
		/// sendで返ってきた値がOKだったかどうかを調べる
		/// </summary>
		/// <param name="retVal">sendの返却値</param>
		/// <returns>trueの場合OKな返却値</returns>
		public bool isOk(string retVal) {
			if (retVal != null) {
				if (retVal.Trim().EndsWith("ok")) {
					return true;
				}
				else if (retVal.Trim().EndsWith("ok.")) {
					return true;
				}
			}
			return false;
		}

		#region マシンの状態に関する情報
		public bool limitX {
			get;
			private set;
		}
		public bool limitY {
			get;
			private set;
		}
		public bool limitZ {
			get;
			private set;
		}
		public enum machineStatus {
			Ready,
			Busy,
			Building,
		}
		public machineStatus status {
			get;
			private set;
		}
		public int currentTempBed {
			get;
			private set;
		}
		public int currentTempNozel {
			get;
			private set;
		}
		public int targetTempBed {
			get;
			private set;
		}
		public int targetTempNozel {
			get;
			private set;
		}
		public int sdProgress {
			get;
			private set;
		}
		public int sdMax {
			get;
			private set;
		}
		public string getStatus() {
			return string.Format("ノズル {0}/{1}, ベッド {2}/{3}, 機器 {4}, 印刷 {5}/{6}", currentTempNozel, targetTempNozel, currentTempBed, targetTempBed, status.ToString(), sdProgress, sdMax);
		}
		#endregion
		/// <summary>
		/// Adventurer3の情報の取り出し(更新)
		/// 機器の状態・温度・JOB状態を取り出す
		/// </summary>
		public void updateStatus() {
			// 機器の状態取得
			var work = send("M119");
			if (isOk(work)) {
				var split = work.Split('\n');
				foreach(var line in split) {
					var trimLine = line.Trim();
					if (trimLine.StartsWith("Endstop")) {
						var endstop = trimLine.Split(' ');
						if (endstop.Length == 4) {
							limitX = limitY = limitZ = true;
							if (endstop[1].EndsWith("0")) {
								limitX = false;
							}
							if (endstop[2].EndsWith("0")) {
								limitY = false;
							}
							if (endstop[3].EndsWith("0")) {
								limitZ = false;
							}
						}
					}
					else if (trimLine.StartsWith("MachineStatus")) {
						if (trimLine.EndsWith("READY")) {
							status = machineStatus.Ready;
						}
						else if (trimLine.EndsWith("BUILDING_FROM_SD")) {
							status = machineStatus.Building;
						}
						else {
							status = machineStatus.Busy;
						}
					}
				}
			}
			// 温度の取得
			work = send("M105");
			if (isOk(work)) {
				var split = work.Split('\n');
				if (split.Length >= 3) {
					char[] delimitter = { ':', '/', 'B' };
					var splitLine = split[1].Trim().Split(delimitter);
					if (splitLine.Length == 6) {
						int temp;
						int.TryParse(splitLine[1], out temp);
						currentTempNozel = temp;
						int.TryParse(splitLine[2], out temp);
						targetTempNozel = temp;
						int.TryParse(splitLine[4], out temp);
						currentTempBed = temp;
						int.TryParse(splitLine[5], out temp);
						targetTempBed = temp;
					}
				}
			}
			// JOB状態の取得
			work = send("M27");
			if (isOk(work)) {
				var split = work.Split('\n');
				if (split.Length >= 3) {
					char[] delimitter = { ' ', '/' };
					var splitLine = split[1].Trim().Split(delimitter);
					if (splitLine.Length == 5) {
						int num;
						int.TryParse(splitLine[3], out num);
						sdProgress = num;
						int.TryParse(splitLine[4], out num);
						sdMax = num;
					}
				}
			}
		}

		/// <summary>
		/// LEDの表示・消去
		/// </summary>
		/// <param name="on">LEDの状態</param>
		public void led(bool on) {
			if (on) {
				send("M146 r255 g255 b255 F0");
			}
			else {
				send("M146 r0 g0 b0 F0");
			}
		}

		static void getByte(int value, byte[] buffer, uint index) {
			buffer[index] = (byte)((value >> 24) & 0xff);
			buffer[index + 1] = (byte)((value >> 16) & 0xff);
			buffer[index + 2] = (byte)((value >> 8) & 0xff);
			buffer[index + 3] = (byte)(value & 0xff);
		}

		private bool isStop_ = false;
		private long remain_;
		/// <summary>
		/// 残りの転送データ量
		/// </summary>
		public long remainTransferByte {
			get { return remain_; }
		}
		/// <summary>
		/// 転送のストップ
		/// </summary>
		public void stopTransfer() {
			isStop_ = true;
		}
		const int blockSize_ = 4096;
		const int headerSize_ = 16;
		/// <summary>
		/// JOB開始
		/// </summary>
		/// <param name="filename">出力するファイル</param>
		/// <param name="jobName">
		/// Adventurer3側に表示するファイル名
		/// これを一応JOB名としておく
		/// 基本的には元のファイル名が望ましい
		/// </param>
		/// <remarks>
		/// Sleep100を入れた理由は、機器側からのブロック送信が完了した通信が再送されたため、送受信データの段ずれが起こった。
		/// プロトコルアナライザの結果を見る限り、以下のような送受信だった。
		/// ・PCから0ブロック目送信
		/// ・機器から0ブロックOK受信
		/// ・PCから1ブロック目送信
		/// ・機器から0ブロックOK受信　← これがおかしい
		/// ・PCから2ブロック目送信
		/// PC側からのSEQ/ACKの番号は問題ないようだったのだが、何らかの理由で、PCからの1ブロック目送信のSEQ/ACKが認識できなかったのかも。
		/// ただ、送信されたファイルで正しく出力はされたみたいだった。
		/// ここで、sleepを入れたところ、とりあえず、送受信の段ずれは解消されたが、機器からの0ブロック目OK受信の前に、なにやら機器側から意味不明な送信が来ている。これは未解消。
		/// </remarks>
		public bool startJob(string filename, string jobName = null) {
			if (jobName == null) {
				// job名が設定されていない場合は、filenameから引用する
				jobName = Path.GetFileName(filename);
			}
			using (var fp = new FileStream(filename, FileMode.Open, FileAccess.Read)) {
				if (isOk(send("M28 " + fp.Length.ToString() + " 0:/user/" + jobName))) {
					int counter = 0;
					isStop_ = false;
					remain_ = fp.Length;
					var crc32 = new Crc32b();	// 検査用ハッシュ関数
					for (long i = 0; i < fp.Length && !isStop_; i += blockSize_, remain_ -= blockSize_, counter++) {
						var readBuffer = new byte[blockSize_];
						var readByte = fp.Read(readBuffer, 0, blockSize_);
						if (readByte <= 0) {
							// 読み取りデータがないので、中止
							return false;
						}
						Array.Resize(ref readBuffer, readByte);
						var writeBuffer = new byte[headerSize_ + blockSize_];
						writeBuffer[0] = writeBuffer[1] = 0x5a;
						writeBuffer[2] = writeBuffer[3] = 0xa5;
						getByte(counter, writeBuffer, 4);
						getByte(readByte, writeBuffer, 8);
						Array.Copy(crc32.ComputeHash(readBuffer), 0, writeBuffer, 12, 4);
						Array.Copy(readBuffer, 0, writeBuffer, 16, readBuffer.Length);
						var ret = send(writeBuffer);
						if (!isOk(ret)) {
							// データの送信ができなかったので中止
							return false;
						}
						Thread.Sleep(100);
					}
				}
				else {
					return false;
				}
			}
			if (!isStop_) {
				isStop_ = false;
				if (isOk(send("M29"))) {
					if (isOk(send("M23 " + "0:/user/" + jobName))) {
						return true;
					}
				}
			}
			return false;
		}

		/// <summary>
		/// JOB中止
		/// </summary>
		public void stopJob() {
			if (isStop_) {
				stopTransfer();
			}
			else {
				send("M26");
			}
		}

		/// <summary>
		/// 機器の緊急停止
		/// </summary>
		public void stop() {
			if (isStop_) {
				stopTransfer();
			}
			else {
				send("M112");
			}
		}
	}
}
