﻿using System;
using System.IO;
using System.Runtime.Serialization;
using System.Collections.Generic;
using System.Text;

namespace AControler {
	/// <summary>
	/// ファイルを更新するためのパラメータ
	/// </summary>
	[DataContract]
	public class FileModifyParameter {
		[DataContract]
		public enum BrimSpeedType {
			NoChange = 0,
			Absolute = 1,
			Ratio = 2
		}
		[DataMember]
		public bool enclosureFanOn { set; get; }
		[DataMember]
		public byte motorX { set; get; }
		[DataMember]
		public byte motorY { set; get; }
		[DataMember]
		public byte motorZ { set; get; }
		[DataMember]
		public byte motorA { set; get; }
		[DataMember]
		public byte motorB { set; get; }
		[DataMember]
		public double playRemovalLength { set; get; }
		[DataMember]
		public double offsetZ { set; get; }
		[DataMember]
		public BrimSpeedType brimSpeedType { set; get; }
		[DataMember]
		public int brimSpeed { set; get; }
		[DataMember]
		public int brimSpeedRatio { set; get; }
		[DataMember]
		public int brimExtrudeRatio { set; get; }

		public FileModifyParameter() {
			enclosureFanOn = false;
			motorX = motorY = motorA = 100;
			motorZ = 40;
			motorB = 20;
			playRemovalLength = 0.5;
			offsetZ = 0.0;
			brimSpeed = 420;
			brimSpeedRatio = 50;
			brimExtrudeRatio = 20;
			brimSpeedType = BrimSpeedType.NoChange;
		}

		public void set(FileModifyParameter source) {
			enclosureFanOn = source.enclosureFanOn;
			motorX = source.motorX;
			motorY = source.motorY;
			motorZ = source.motorZ;
			motorA = source.motorA;
			motorB = source.motorB;
			playRemovalLength = source.playRemovalLength;
			offsetZ = source.offsetZ;
			brimSpeedType = source.brimSpeedType;
			brimSpeed = source.brimSpeed;
			brimSpeedRatio = source.brimSpeedRatio;
			brimExtrudeRatio = source.brimExtrudeRatio;
		}
	}

	/// <summary>
	/// ファイル編集クラス
	/// </summary>
	public class ModifyFile {
		/// <summary>
		/// XとYの値が一致しているかどうかを判断する。
		/// 根拠とするε値は、小数点4桁まで判断可能な値にする
		/// </summary>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <returns></returns>
		static bool isEqual(Double x, Double y) {
			return Math.Abs(x - y) < 0.0001;
		}

		/// <summary>
		/// 文字列を実数に変換し、前回値の更新を行う
		/// 一応、内部的に保持している絶対値・相対値情報を使う
		/// </summary>
		/// <param name="prev">前の値</param>
		/// <param name="value">今回の値</param>
		/// <returns>計算結果</returns>
		static Double moveValue(Double prev, string value, bool isAbsolute) {
			Double work;
			if (Double.TryParse(value, out work)) {
				if (isAbsolute) {
					return work;
				}
				else {
					return prev + work;
				}
			}
			return prev;
		}

		/// <summary>
		/// ファイル種別
		/// </summary>
		public enum FileType {
			FlashPrint,
			Simplify3D,
			Other,
		}

		/// <summary>
		/// Simplify3Dで出力されたファイル内のパラメータ。
		/// </summary>
		class Simplify3DParameter {
			public int rapidXYspeed_ = 4800;
			public int rapidZspeed_ = 300;
			public int defaultSpeed_ = 3600;

			/// <summary>
			/// パラメータ行の解析
			/// </summary>
			/// <param name="value">解析対象の1行の情報</param>
			/// <returns>パラメータとして落とし込めたらtrue</returns>
			public bool parseParameter(string value) {
				var lines = value.Trim().Split(',');
				if (lines.Length >= 2) {
					if (lines[0].StartsWith(";   defaultSpeed")) {
						return int.TryParse(lines[1], out defaultSpeed_);
					}
					else if (lines[0].StartsWith(";   rapidXYspeed")) {
						return int.TryParse(lines[1], out rapidXYspeed_);
					}
					else if (lines[0].StartsWith(";   rapidZspeed")) {
						return int.TryParse(lines[1], out rapidZspeed_);
					}
				}
				return false;
			}
		}

		/// <summary>
		/// 1行前までのX,Y,Z,Eの値、その他Gコードの情報
		/// </summary>
		class GValue {
			public Double X_;
			public Double Y_;
			public Double Z_;
			public Double E_;
			public int F_;

			public GValue() {
				X_ = Y_ = Z_ = E_ = 0.0;
				F_ = 0;
			}
			public GValue(GValue target) {
				set(target);
			}
			public void set(GValue target) {
				X_ = target.X_;
				Y_ = target.Y_;
				Z_ = target.Z_;
				E_ = target.E_;
				F_ = target.F_;
			}
		}

		/// <summary>
		/// 現在処理しているファイルの種別
		/// </summary>
		FileType fileType_ = FileType.Other;

		Stream inputFile_;
		Stream outputFile_;
		Simplify3DParameter s3dParameter_;
		FileModifyParameter modifyParameter_ = new FileModifyParameter();
		ASCIIEncoding ascii_ = new ASCIIEncoding();

		/// <summary>
		/// パラメータの入出力
		/// </summary>
		public FileModifyParameter parameter {
			get {
				return modifyParameter_;
			}
			set {
				modifyParameter_.set(value);
			}
		}

		string readLine() {
			var buffer = new List<Byte>();
			for (var val = inputFile_.ReadByte(); val != -1; val = inputFile_.ReadByte()) {
				buffer.Add((Byte)val);
				if (val == 0x0a) {
					break;
				}
			}
			return ascii_.GetString(buffer.ToArray());
		}

		void write(string data) {
			var work = ascii_.GetBytes(data);
			outputFile_.Write(work, 0, work.Length);
		}

		void writeLine(string data) {
			write(data);
			if (fileType_ == FileType.Simplify3D) {
				outputFile_.WriteByte(0x0d);
			}
			outputFile_.WriteByte(0x0a);
		}

		bool endOfStream() {
			return inputFile_.Position >= inputFile_.Length;
		}

		/// <summary>
		/// ファイルの種別を調べて、必要であれば、前処理を行う。
		/// </summary>
		/// <returns>trueの場合、処理可能なファイル</returns>
		bool checkFileType() {
			// 1行目を読み込み、ファイル種別を取得する
			var firstLine = readLine();
			if (firstLine.StartsWith("xgcode 1.0")) {
				fileType_ = FileType.FlashPrint;
				write(firstLine);
				// ;start gcodeまでそのまま出力する。
				int headerSize;
				{
					// まずヘッダーに関する情報を取得、出力する。
					var work = new byte[0x20 - firstLine.Length];
					var size = inputFile_.Read(work, 0, work.Length);
					outputFile_.Write(work, 0, work.Length);
					headerSize = BitConverter.ToInt32(work, 0x14 - firstLine.Length);
				}
				{
					var work = new byte[headerSize - 0x20];
					var size = inputFile_.Read(work, 0, work.Length);
					outputFile_.Write(work, 0, size);
				}
				return true;
			}
			else if (firstLine.StartsWith("; G-Code generated by Simplify3D")) {
				fileType_ = FileType.Simplify3D;
				return true;
			}
			return false;
		}

		/// <summary>
		/// 文字列の中から英文字(1文字)をコマンドとして、その後に続く数値と.と-を値として解析して返す
		/// </summary>
		/// <param name="line">ファイル中の1行の情報</param>
		/// <param name="command">コマンド文字を小文字にしたもの</param>
		/// <param name="value">数値らしき値</param>
		/// <returns>commandとvalueを除いた値</returns>
		string parseCommand(string line, out char command, out string value) {
			var work = line.Trim();
			command = ' ';
			value = "";
			if (work.Length > 0) {
				command = work[0];
				int next = 1;
				for (; next < work.Length && (Char.IsDigit(work[next]) || work[next] == '-' || work[next] == '.'); next++) {
					value += work[next];
				}
				work = work.Substring(next, work.Length - next);
			}
			return work;
		}

		/// <summary>
		/// Simplify3D形式で、ファイルの先頭側で出力する項目
		/// </summary>
		void writePre() {
			writeLine("G28");
			writeLine("M132 X Y Z A B");
			writeLine(string.Format("G1 Z50.000 F{0}", s3dParameter_.rapidZspeed_));
			writeLine(string.Format("G161 X Y F{0}", s3dParameter_.rapidXYspeed_));
			writeLine("M7 T0");
			writeLine("M6 T0");
			if (modifyParameter_.enclosureFanOn) {
				writeLine("M651");
			}
			writeLine(string.Format("M907 X{0} Y{1} Z{2} A{3} B{4}", modifyParameter_.motorX, modifyParameter_.motorY, modifyParameter_.motorZ, modifyParameter_.motorA, modifyParameter_.motorB));
		}

		/// <summary>
		/// Simplify3D形式で、ファイルの終了側で出力する項目
		/// </summary>
		void writePost() {
			writeLine("M104 S0 T0");
			writeLine("M140 S0 T0");
			writeLine(string.Format("G162 Z F{0}", s3dParameter_.rapidZspeed_));
			writeLine("M107");
			writeLine("G28 X Y");
			if (modifyParameter_.enclosureFanOn) {
				writeLine("M652");
			}
			writeLine("M132 X Y Z A B");
			writeLine("G91");
			writeLine("M18");
		}

		/// <summary>
		/// ブリム除去コマンドの出力
		/// </summary>
		void writePlayRemoval(int rapidZspeed) {
			if (modifyParameter_.playRemovalLength != 0.0) {
				writeLine("G91");
				writeLine(string.Format("G1 Z{0:F1} F{1}", modifyParameter_.playRemovalLength, rapidZspeed));
				writeLine("G90");
			}
		}

		/// <summary>
		/// ファイルの更新処理を実施する
		/// </summary>
		/// <returns>更新処理が正しく完了した場合true</returns>
		bool updateFile() {
			var prev = new GValue();
			bool isAbsolute = true;
			bool brimSection = false;
			prev.Z_ = 10000.0;  // 大きな値をとりあえず設定しておく
			bool isPreOutputed = false;

			while (!endOfStream()) {
				var line = readLine();
				if (line.StartsWith(";")) {
					// コメント行の解析
					if (fileType_ == FileType.Simplify3D) {
						if (!s3dParameter_.parseParameter(line)) {
							// 解釈できなものだった場合の処理
							if (line.StartsWith("; layer end")) {
								// ポスト処理を実施する
								writePost();
							}
							else if (line.StartsWith("; process") && !isPreOutputed) {
								// プリ処理を実施する
								writePre();
								isPreOutputed = true;
							}
							else if (line.StartsWith("; feature skirt")) {
								// ブリム開始位置
								brimSection = true;
							}
							else if (line.StartsWith("; feature")) {
								// ブリム終了位置
								brimSection = false;
							}
						}
					}
					else {
						write(line);
					}
				}
				else {
					var current = new GValue(prev);
					long gMode = 0;
					bool hasZ_ = false;
					bool hasXY_ = false;
					bool hasE_ = false;
					string outputString = "";

					var work = line.Trim();
					if (work.IndexOf(';') >= 0) {
						work = work.Remove(work.IndexOf(';'));
					}
					while (work.Length > 0) {
						char command;
						string value;
						work = parseCommand(work, out command, out value);
						if (command == ' ') {
							// データが読み込まれていない場合なので
							break;
						}
						if (outputString.Length > 0) {
							outputString += " ";
						}
						switch (command) {
						case 'G':
							switch (value) {
							case "1":
								// ヘッド移動モード
								gMode = 1;
								break;
							case "91":
								// 相対移動
								isAbsolute = false;
								break;
							case "90":
								// 絶対移動
								isAbsolute = true;
								break;
							}
							outputString += command + value;
							break;
						case 'X':
						case 'Y':
							hasXY_ = true;
							outputString += command + value;
							break;
						case 'Z':
							hasZ_ = true;
							current.Z_ = moveValue(prev.Z_, value, isAbsolute);
							if (modifyParameter_.offsetZ != 0 && isAbsolute && gMode == 1) {
								// グローバルオフセットの対応
								value = (current.Z_ + modifyParameter_.offsetZ).ToString("F3");
							}
							outputString += command + value;
							break;
						case 'E':
							current.E_ = moveValue(prev.E_, value, isAbsolute);
							hasE_ = true;
							if (brimSection && gMode == 1 && hasXY_ && modifyParameter_.brimExtrudeRatio != 100) {
								// ブリムの吐出量の変更G1 X? Y? の場合
								outputString += command + (current.E_ * (double)modifyParameter_.brimExtrudeRatio / 100.0).ToString("F4");
							}
							else {
								outputString += command + value;
							}
							break;
						case 'F':
							current.F_ = (int)moveValue(0.0, value, true);
							if (brimSection && gMode == 1 && hasXY_ && hasE_ && modifyParameter_.brimSpeedType != FileModifyParameter.BrimSpeedType.NoChange) {
								// ブリムの吐出速度を変更
								if (modifyParameter_.brimSpeedType == FileModifyParameter.BrimSpeedType.Ratio) {
									outputString += command + (current.F_ * (double)modifyParameter_.brimSpeedRatio / 100.0).ToString();
								}
								else {
									outputString += command + modifyParameter_.brimSpeed.ToString();
								}
							}
							else {
								outputString += command + value;
							}
							break;
						default:
							outputString += command + value;
							break;
						}
					}
					if (gMode == 1 && hasZ_ && current.Z_ > prev.Z_) {
						// G1かつZ値がある場合遊び除去コマンドを事前に出力
						writePlayRemoval(fileType_ == FileType.FlashPrint ? current.F_: s3dParameter_.rapidZspeed_);
					}
					writeLine(outputString);
					if (gMode == 1) {
						prev.set(current);
					}
				}
			}
			return true;
		}

		/// <summary>
		/// ファイルの更新処理
		/// </summary>
		/// <param name="inputFile">入力ファイル</param>
		/// <param name="outputFile">出力ファイル</param>
		/// <returns>更新が正しく終了した場合true</returns>
		public bool modify(Stream inputFile, Stream outputFile) {
			inputFile_ = inputFile;
			outputFile_ = outputFile;
			s3dParameter_ = new Simplify3DParameter();
			if (checkFileType()) {
				return updateFile();
			}
			return false;
		}
	}
}
