﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.IO;
using AControler;
using System.Runtime.Serialization.Json;

namespace PrintConsole {
	class Program {
		static void Main(string[] args) {
			var work = new Program();
			if (work.connect()) {
				work.printInfomation();
				work.startUpdate();
				work.ui();
			}
		}

		Task sync_;
		const int statusCheckInterval_ = 3000;
		private AControler.Adventurer3 target_;
		/// <summary>
		/// Adventurer3へのコマンド送信がバッティングしないようにするための排他制御用
		/// </summary>
		private Mutex mut_ = new Mutex();

		/// <summary>
		/// Adventurer3との接続
		/// </summary>
		/// <returns>接続できた</returns>
		bool connect() {
			var list = AControler.Adventurer3.searchIP();
			if (list.Count > 0) {
				target_ = new Adventurer3(list[0]);
				target_.start();
				return target_.isConnected;
			}
			else {
				return false;
			}
		}

		void printInfomation() {
			var ret = target_.send("M115");
			if (target_.isOk(ret)) {
				Console.WriteLine(string.Format("Connected {0}", target_.ip));
				Console.WriteLine(ret);
			}
			else {
				Console.WriteLine("Error");
			}
		}

		/// <summary>
		/// ステータス更新用のバックグラウンド処理の開始
		/// </summary>
		void startUpdate() {
			if (sync_ != null) {
				if (sync_.IsCanceled || sync_.IsCompleted) {
					sync_.Dispose();
					sync_ = null;
				}
			}
			if (sync_ == null) {
				sync_ = Task.Run(() => statusUpdate());
			}
		}

		/// <summary>
		/// 1秒単位にステータス更新
		/// </summary>
		void statusUpdate() {
			while(target_.isConnected) {
				if (mut_.WaitOne(0)) {
					// ロック中の場合、データ送信されていることになるので、今回の更新はパスする。
					target_.updateStatus();
					mut_.ReleaseMutex();
				}
				Thread.Sleep(statusCheckInterval_);
			}
		}

		/// <summary>
		/// 印刷コマンド
		/// </summary>
		/// <param name="command">コマンド名そのまま</param>
		void print(string command) {
			// コマンドをとりあえずスペースでパースする
			var cmds = command.Split(' ');
			var modifier = new ModifyFile();
			var filename = cmds[0].Substring(1);
			if (cmds.Length >= 2 && cmds[1].EndsWith(".json")) {
				if (File.Exists(cmds[1])) {
					using (var wStream = File.OpenRead(cmds[1])) {
						var settings = new DataContractJsonSerializerSettings();
						var serializer = new DataContractJsonSerializer(typeof(FileModifyParameter), settings);
						var data1 = (FileModifyParameter)serializer.ReadObject(wStream);
						modifier.parameter.set(data1);
					}
					bool fileModifyResult = false;
					var tempFilename = Path.GetTempFileName();
					using (var inFile = File.OpenRead(filename)) {
						using (var outFile = File.Create(tempFilename)) {
							fileModifyResult = modifier.modify(inFile, outFile);
						}
					}
					if (fileModifyResult) {
						if (mut_.WaitOne()) {
							var ret = target_.startJob(tempFilename, Path.GetFileNameWithoutExtension(filename));
							mut_.ReleaseMutex();
							if (!ret) {
								Console.WriteLine("Error");
							}
							else {
								Console.WriteLine("Start Job");
							}
						}
						File.Delete(tempFilename);
					}
				}
			}
			else {
				if (mut_.WaitOne()) {
					var ret = target_.startJob(filename);
					mut_.ReleaseMutex();
					if (!ret) {
						Console.WriteLine("Error");
					}
					else {
						Console.WriteLine("Start Job");
					}
				}
			}
		}

		/// <summary>
		/// UI操作
		/// </summary>
		void ui() {
			for (; ;) {
				Console.Write("> ");
				var cmd = Console.ReadLine();
				if (!target_.isConnected) {
					target_.start();
					if (!target_.isConnected) {
						// 再接続できなかったので終了
						break;
					}
					startUpdate();
				}
				// コマンド処理
				cmd = cmd.Trim();
				if (cmd.StartsWith("q") || cmd.StartsWith("Q")) {
					// コマンド中止
					target_.end();
					break;
				}
				if (cmd.StartsWith("p") || cmd.StartsWith("P")) {
					Console.WriteLine(target_.getStatus());
				}
				else if (cmd.StartsWith("l") || cmd.StartsWith("L")) {
					if (mut_.WaitOne()) {
						if (cmd.EndsWith("0")) {
							target_.led(false);
						}
						else {
							target_.led(true);
						}
						mut_.ReleaseMutex();
					}
				}
				else if (cmd.StartsWith("s") || cmd.StartsWith("S")) {
					if (mut_.WaitOne()) {
						target_.stop();
						mut_.ReleaseMutex();
						Console.WriteLine("Stop");
					}
				}
				else if (cmd.StartsWith("G") || cmd.StartsWith("M")) {
					if (mut_.WaitOne()) {
						var ret = target_.send(cmd);
						if (ret != null) {
							Console.WriteLine(ret);
						}
						else {
							// データ送受信失敗なので、再接続をする
							Console.WriteLine("Error");
							target_.end();
							target_.start();
							startUpdate();
						}
						mut_.ReleaseMutex();
					}
				}
				else if (cmd.StartsWith("jobstop")) {
					if (mut_.WaitOne()) {
						target_.stopJob();
						mut_.ReleaseMutex();
						Console.WriteLine("Stop Job");
					}
				}
				else if (cmd.StartsWith("j") || cmd.StartsWith("J")) {
					print(cmd);
				}
			}
		}

		~Program() {
			mut_.Dispose();
		}
	}
}
